#ifdef SPARC64

#include "defs.h"
#include <stdio.h>
#include <elf.h>
#include <asm/ptrace.h>
#include <linux/const.h>

/* Things not done, debugged or tested at this point:
 * 1) uvtop swap handling
 * 2) uniform page table layout - like we had in 1st quarter of 2013
 * 3) and whatever can't be thought of.
 */
#define	true	(1)
#define	false	(0)

static const unsigned long not_valid_pte = ~0UL;
static struct machine_specific sparc64_machine_specific;
static unsigned long sparc64_ksp_offset;
static unsigned long vmemmap_table;
static int two_level_vmemmap_table;
#define VMEMMAP_CHUNK_SHIFT	(22)
#define VMEMMAP_CHUNK		(1UL << VMEMMAP_CHUNK_SHIFT)
#define VMEMMAP_CHUNK_MASK	(~(VMEMMAP_CHUNK - 1UL))
#define MAX_PHYS_ADDRESS_LOBITS	(41)
#define NR_CHUNKS_SHIFT		(MAX_PHYS_ADDRESS_LOBITS - PAGE_SHIFT + 6)
#define	NR_CHUNKS_MASK		(~((1UL << NR_CHUNKS_SHIFT) - 1))
#define	MAGIC_TT		(0x1ff)

static unsigned long __va(unsigned long paddr)
{
	return paddr + PAGE_OFFSET;
}

static unsigned long __pa(unsigned long vaddr)
{
	return vaddr - PAGE_OFFSET;
}

static uint page_size(void)
{
	return machdep->pagesize;
}

static void sparc64_parse_cmdline_args(void)
{
}

/* This interface might not be required. */
static void sparc64_clear_machdep_cache(void)
{
}

/*
 *  "mach" command output.
 */
static void
sparc64_display_machine_stats(void)
{
	int c;
	struct new_utsname *uts;
	char buf[BUFSIZE];
	ulong mhz;

	uts = &kt->utsname;

	fprintf(fp, "          MACHINE TYPE: %s\n", uts->machine);
	fprintf(fp, "           MEMORY SIZE: %s\n", get_memory_size(buf));
	fprintf(fp, "                  CPUS: %d\n", kt->cpus);
	fprintf(fp, "       PROCESSOR SPEED: ");
	if ((mhz = machdep->processor_speed()))
		fprintf(fp, "%ld Mhz\n", mhz);
	else
		fprintf(fp, "(unknown)\n");
	fprintf(fp, "                    HZ: %d\n", machdep->hz);
	fprintf(fp, "             PAGE SIZE: %d\n", PAGESIZE());
	fprintf(fp, "   KERNEL VIRTUAL BASE: %lx\n", machdep->kvbase);
	fprintf(fp, "   KERNEL VMALLOC BASE: %lx\n", vt->vmalloc_start);
	if (machdep->flags & VMEMMAP)
		fprintf(fp, "   KERNEL VMEMMAP BASE: %lx\n",
			machdep->machspec->vmemmap_vaddr);
	fprintf(fp, "   KERNEL MODULES BASE: %lx\n", MODULES_VADDR);
	fprintf(fp, "     KERNEL STACK SIZE: %ld\n", STACKSIZE());

	fprintf(fp, "HARD IRQ STACK SIZE: %ld\n", THREAD_SIZE);
	fprintf(fp, "    HARD IRQ STACKS:\n");

	for (c = 0; c < kt->cpus; c++) {
		if (!tt->hardirq_ctx[c])
			continue;
		sprintf(buf, "CPU %d", c);
		fprintf(fp, "%19s: %lx\n", buf, tt->hardirq_ctx[c]);
	}

	fprintf(fp, "SOFT IRQ STACK SIZE: %ld\n", THREAD_SIZE);
	fprintf(fp, "    SOFT IRQ STACKS:\n");
	for (c = 0; c < kt->cpus; c++) {
		if (!tt->softirq_ctx[c])
			continue;
		sprintf(buf, "CPU %d", c);
		fprintf(fp, "%19s: %lx\n", buf, tt->softirq_ctx[c]);
	}
}

static void sparc64_display_memmap(void)
{
	unsigned long iomem_resource;
	unsigned long resource;
	unsigned long start, end, nameptr;
	int size = STRUCT_SIZE("resource");
	char *buf = GETBUF(size);
	char name[32];

	iomem_resource = symbol_value("iomem_resource");

	readmem(iomem_resource + MEMBER_OFFSET("resource", "child"), KVADDR,
		&resource, sizeof(resource), "iomem_resource", FAULT_ON_ERROR);

	fprintf(fp, "      PHYSICAL ADDRESS RANGE         TYPE\n");

	while (resource) {
		readmem(resource, KVADDR, buf, size, "resource",
			FAULT_ON_ERROR);
		start = ULONG(buf + MEMBER_OFFSET("resource", "start"));
		end = ULONG(buf + MEMBER_OFFSET("resource", "end"));
		nameptr = ULONG(buf + MEMBER_OFFSET("resource", "name"));

		readmem(nameptr, KVADDR, name, sizeof(name), "resource.name",
			FAULT_ON_ERROR);

		fprintf(fp, "%016lx - %016lx  %-32s\n", start, end, name);

		resource = ULONG(buf + MEMBER_OFFSET("resource", "sibling"));
	}
}

static void sparc64_cmd_mach(void)
{
	int c;
	int mflag = 0;

	while ((c = getopt(argcnt, args, "cm")) != EOF) {
		switch (c) {
		case 'm':
			mflag++;
			sparc64_display_memmap();
			break;
		case 'c':
			fprintf(fp, "SPARC64: '-%c' option is not supported\n",
				c);
			break;
		default:
			argerrs++;
			break;
		}
	}

	if (argerrs)
		cmd_usage(pc->curcmd, SYNOPSIS);

	if (!mflag)
		sparc64_display_machine_stats();
}

struct sparc64_mem_ranges {
	unsigned long start;
	unsigned long end;
};

#define NR_PHYS_RANGES	(128)
static unsigned int nr_phys_ranges;
struct sparc64_mem_ranges phys_ranges[NR_PHYS_RANGES];

#define NR_IMAGE_RANGES (16)
static unsigned int nr_kimage_ranges;
struct sparc64_mem_ranges kimage_ranges[NR_IMAGE_RANGES];

/* There are three live cases:
 *  one) normal kernel
 *  two) --load-panic kernel
 *  and
 *  three) --load kernel
 * One and two can be treated the same because the kernel is physically
 * contiguous. Three isn't contiguous. The kernel is allocated in order
 * nine allocation pages. We don't handle case three yet.
 */

static int sparc64_phys_live_valid(unsigned long paddr)
{
	unsigned int nr;
	int rc = false;

	for (nr = 0U; nr != nr_phys_ranges; nr++) {
		if (paddr >= phys_ranges[nr].start &&
			paddr < phys_ranges[nr].end) {
			rc = true;
			break;
		}
	}
	return rc;
}

static int sparc64_phys_kdump_valid(unsigned long paddr)
{
	return true;
}

static int sparc64_verify_paddr(unsigned long paddr)
{
	int rc;

	if (ACTIVE())
		rc = sparc64_phys_live_valid(paddr);
	else
		rc = sparc64_phys_kdump_valid(paddr);

	return rc;
}

static void sparc6_phys_base_live_limits(void)
{
	if (nr_phys_ranges >= NR_PHYS_RANGES)
		error(FATAL, "sparc6_phys_base_live_limits: "
			"NR_PHYS_RANGES exceeded.\n");
	else if (nr_kimage_ranges >= NR_IMAGE_RANGES)
		error(FATAL, "sparc6_phys_base_live_limits: "
			"NR_IMAGE_RANGES exceeded.\n");
}

static void sparc64_phys_base_live_valid(void)
{
	if (!nr_phys_ranges)
		error(FATAL, "No physical memory ranges.");
	else if (!nr_kimage_ranges)
		error(FATAL, "No vmlinux memory ranges.");
}

static void sparc64_phys_base_live(void)
{
	char line[BUFSIZE];
	FILE *fp;

	fp = fopen("/proc/iomem", "r");
	if (fp == NULL)
		error(FATAL, "Can't open /proc/iomem. We can't proceed.");

	while (fgets(line, sizeof(line), fp) != 0) {
		unsigned long start, end;
		int count, consumed;
		char *ch;

		sparc6_phys_base_live_limits();
		count = sscanf(line, "%lx-%lx : %n", &start, &end, &consumed);
		if (count != 2)
			continue;
		ch = line + consumed;
		if (memcmp(ch, "System RAM\n", 11) == 0) {
			end = end + 1;
			phys_ranges[nr_phys_ranges].start = start;
			phys_ranges[nr_phys_ranges].end = end;
			nr_phys_ranges++;
		} else if ((memcmp(ch, "Kernel code\n", 12) == 0) ||
				(memcmp(ch, "Kernel data\n", 12) == 0) ||
				(memcmp(ch, "Kernel bss\n", 11) == 0)) {
			kimage_ranges[nr_kimage_ranges].start = start;
			kimage_ranges[nr_kimage_ranges].end = end;
			nr_kimage_ranges++;
		}
	}

	(void) fclose(fp);
	sparc64_phys_base_live_valid();
}

static void sparc64_phys_base_kdump(void)
{
}

static void sparc64_phys_base(void)
{
	if (ACTIVE())
		return sparc64_phys_base_live();
	else
		return sparc64_phys_base_kdump();
}

static unsigned long kva_start, kva_end;
static unsigned long kpa_start, kpa_end;

static void sparc64_kimage_limits_live(void)
{
	kpa_start = kimage_ranges[0].start;
	kpa_end = kpa_start + (kva_end - kva_start);
}

static void sparc64_kimage_limits_kdump(void)
{
	unsigned long phys_base;

	if (DISKDUMP_DUMPFILE()) {
		if (diskdump_phys_base(&phys_base)) {
			kpa_start = phys_base | (kva_start & 0xffff);
			kpa_end = kpa_start + (kva_end - kva_start);
			return;
		}
	}
	fprintf(stderr, "Can't determine phys_base\n");
}

static unsigned long kimage_va_translate(unsigned long addr)
{
	unsigned long paddr = (addr - kva_start) + kpa_start;

	return paddr;
}

static int kimage_va_range(unsigned long addr)
{
	if (addr >= kva_start && addr < kva_end)
		return true;
	else
		return false;
}

static void sparc64_kimage_limits(void)
{
	kva_start = symbol_value("_stext");
	kva_end = symbol_value("_end");

	if (ACTIVE())
		sparc64_kimage_limits_live();
	else
		sparc64_kimage_limits_kdump();
}

static int sparc64_is_linear_mapped(unsigned long vaddr)
{
	int rc = 0;

	if  ((vaddr & PAGE_OFFSET) == PAGE_OFFSET)
		rc = 1;
	return rc;
}

static unsigned long pte_to_pa(unsigned long pte)
{
	unsigned long paddr = pte & _PAGE_PFN_MASK;

	return paddr;
}

static unsigned long sparc64_vmemmap_translate(unsigned long vaddr)
{
	struct machine_specific *ms = &sparc64_machine_specific;
	unsigned long offset = vaddr - ms->vmemmap_vaddr;
	unsigned long chunk_offset = offset & ~VMEMMAP_CHUNK_MASK;
	unsigned long chunk;
	unsigned long kva;
	unsigned long paddr = 0UL;
	unsigned long pte;
	unsigned long index;
	unsigned long pte_paddr;
	int rc;

	if (two_level_vmemmap_table) {
		index = offset >> NR_CHUNKS_SHIFT;
		kva = vmemmap_table + (index * sizeof(long));
		rc = readmem(kva, KVADDR, &pte_paddr, sizeof(long),
			     "sparc64_vmemmap_translate: error.",
			     RETURN_ON_ERROR);
		if (!rc)
			goto out;
		chunk = (offset & ~NR_CHUNKS_MASK) >> VMEMMAP_CHUNK_SHIFT;
		pte_paddr += chunk * sizeof(pte);
		rc = readmem(pte_paddr, PHYSADDR, &pte, sizeof(pte),
			     "sparc64_vmemmap_translate: error.",
			     RETURN_ON_ERROR);
		if (!rc)
			goto out;
	} else {
		chunk = offset >> VMEMMAP_CHUNK_SHIFT;
		kva = chunk * sizeof(pte);
		kva = kva + vmemmap_table;
		rc = readmem(kva, KVADDR, &pte, sizeof(pte),
			     "sparc64_vmemmap_translate: error.",
			     RETURN_ON_ERROR);
		if (!rc)
			goto out;
	}
	paddr = pte_to_pa(pte);
	paddr = paddr | chunk_offset;
out:
	return paddr;
}

static int sparc64_is_vmemmap_mapped(unsigned long vaddr)
{
	struct machine_specific *ms = &sparc64_machine_specific;
	int rc = false;

	if (ms->vmemmap_vaddr && vaddr >= ms->vmemmap_vaddr &&
	    vaddr < ms->vmemmap_end)
		rc = true;
	return rc;
}

static unsigned long fetch_page_table_level(unsigned long pte_kva,
				unsigned long vaddr, unsigned int shift,
				unsigned int mask, const char *name,
				int verbose)
{
	unsigned int pte_index = (vaddr >> shift) & mask;
	unsigned long page_table[PTES_PER_PAGE];
	unsigned long pte = 0UL;
	int rc;

	rc = readmem(pte_kva, KVADDR, page_table, sizeof (page_table),
		(char *) name, RETURN_ON_ERROR);
	if (!rc)
		goto out;
	pte = page_table[pte_index];
	if (verbose)
		fprintf(fp,
			"%s(0x%.16lx) fetch of pte @index[0x%.4x]=0x%.16lx\n",
			name, pte_kva, pte_index, pte);
out:
	return pte;
}

static unsigned int ptes_per_page_mask(void)
{
	return PTES_PER_PAGE_MASK;
}

static unsigned long pmd_is_huge(unsigned long pmd, unsigned long vaddr,
							int verbose)
{
	struct machine_specific *ms = &sparc64_machine_specific;
	unsigned long hpage_mask;
	unsigned long paddr = 0UL;

	/* We use the pud_shift again to check for uniform versus
	 * non-uniform page table layout.
	 */
	if (ms->pud_shift == PG_UEK3_PUD_SHIFT) {
		if ((pmd & NONUNIFORM_PMD_ISHUGE) == 0UL)
			goto out;
		hpage_mask = ~((1UL << NONUNIFORM_HPAGE_SHIFT) - 1UL);
	} else {
		if ((pmd & UNIFORM_PAGE_PMD_HUGE) == 0UL)
			goto out;
		hpage_mask = ~((1UL << UNIFORM_HPAGE_SHIFT) - 1UL);
	}
	paddr = pte_to_pa(pmd) + (vaddr & ~hpage_mask);
	if (verbose)
		fprintf(fp, "Huge Page/THP pmd=0x%.16lx paddr=0x%.16lx\n",
			pmd, paddr);
out:
	return paddr;
}

static unsigned long sparc64_page_table_walk(unsigned long pgd,
					unsigned long vaddr, int verbose)
{
	struct machine_specific *ms = &sparc64_machine_specific;
	static const char *pgd_text = "pgd fetch";
	static const char *pud_text = "pud fetch";
	static const char *pmd_text = "pmd fetch";
	static const char *pte_text = "pte fetch";
	unsigned long kva = pgd;
	unsigned long paddr;
	unsigned long pte;

	if (!sparc64_is_linear_mapped(kva))
		error(FATAL,
			"sparc64_page_table_walk: pgd must be identity mapped"
			" but isn't (0xlx).", pgd);

	pte = fetch_page_table_level(kva, vaddr, ms->pgd_shift,
		ptes_per_page_mask(), pgd_text, verbose);
	if (!pte)
		goto bad;
	kva = __va(pte);
	/* For PUD folding skip this step. */
	if (ms->pud_shift) {
		pte = fetch_page_table_level(kva, vaddr, ms->pud_shift,
			ptes_per_page_mask(), pud_text, verbose);
		if (!pte)
			goto bad;
	}
	kva = __va(pte);
	pte = fetch_page_table_level(kva, vaddr, ms->pmd_shift,
		ptes_per_page_mask(), pmd_text, verbose);
	if (!pte)
		goto bad;
	/* Check for a huge/THP page */
	paddr = pmd_is_huge(pte, vaddr, verbose);
	if (paddr)
		goto out;
	kva = __va(pte);
	pte = fetch_page_table_level(kva, vaddr, PAGE_SHIFT,
		ms->ptes_per_pte - 1, pte_text, verbose);
	if ((pte & _PAGE_VALID) == 0UL)
		goto bad;
	paddr = pte_to_pa(pte);
	paddr = paddr | (vaddr & ~PAGE_MASK);
out:
	return paddr;
bad:
	return not_valid_pte;
}

static void sparc64_init_kernel_pgd(void)
{
	int cpu, rc;
	ulong v;

	v = symbol_value("init_mm");
	rc = readmem(v + OFFSET(mm_struct_pgd), KVADDR, &v, sizeof (v),
		"init_mm.pgd", RETURN_ON_ERROR);
	if (!rc) {
		error(WARNING, "Can not determine pgd location.\n");
		goto out;
	}

	for (cpu = 0; cpu < NR_CPUS; cpu++)
		vt->kernel_pgd[cpu] = v;
out:
	return;
}

static int sparc64_get_smp_cpus(void)
{
	int ncpu = MAX(get_cpus_online(), get_highest_cpu_online() + 1);

	return ncpu;
}

static void sparc64_init_uek3(void)
{
	struct machine_specific *machspec = &sparc64_machine_specific;

	machspec->userspace_top = PG_UEK3_SPARC64_USERSPACE_TOP;
	machspec->pgd_shift = PG_UEK3_PGDIR_SHIFT;
	machspec->pud_shift = PG_UEK3_PUD_SHIFT;
	machspec->pmd_shift = PG_UEK3_PMD_SHIFT;
	machspec->ptes_per_pte = PG_UEK3_PTRS_PER_PTE;
	machdep->max_physmem_bits = _MAX_PHYSMEM_BITS_UEK3;
}

static void sparc64_init_level4(void)
{
	struct machine_specific *machspec = &sparc64_machine_specific;

	machspec->userspace_top = PG_LVL4_SPARC64_USERSPACE_TOP;
	machspec->pgd_shift = PG_LVL4_PGDIR_SHIFT;
	machspec->pud_shift = PG_LVL4_PUD_SHIFT;
	machspec->pmd_shift = PG_LVL4_PMD_SHIFT;
	machspec->ptes_per_pte = PG_LVL4_PTRS_PER_PTE;
	machdep->max_physmem_bits = _MAX_PHYSMEM_BITS_LVL4;
}

static void sparc64_init_level3(void)
{
	struct machine_specific *machspec = &sparc64_machine_specific;

	machspec->userspace_top = PG_LVL3_SPARC64_USERSPACE_TOP;
	machspec->pgd_shift = PG_LVL3_PGDIR_SHIFT;
	machspec->pud_shift = 0U;
	machspec->pmd_shift = PG_LVL3_PMD_SHIFT;
	machspec->ptes_per_pte = PG_LVL3_PTRS_PER_PTE;
	machdep->max_physmem_bits = _MAX_PHYSMEM_BITS_LVL3;
	error(WARNING, "This hasn't been debugged or tested yet.\n");
}

static ulong sparc64_vmalloc_start(void)
{
	return machdep->machspec->vmalloc_start;
}

int sparc64_IS_VMALLOC_ADDR(ulong vaddr)
{
	int ret = (vaddr >= machdep->machspec->vmalloc_start &&
		vaddr < machdep->machspec->vmalloc_end);

	return ret;
}

static void pt_clear_cache(void)
{
	machdep->last_pgd_read = 0UL;
	machdep->last_pud_read = 0UL;
	machdep->last_pmd_read = 0UL;
	machdep->last_ptbl_read = 0UL;
}

static void pt_level_alloc(char **lvl, char *name)
{
	size_t sz = page_size();
	void *pointer = malloc(sz);

	if (!pointer)
		error(FATAL, name);
	*lvl = pointer;
	return;
}

static int sparc64_verify_symbol(const char *name, unsigned long value,
		char type)
{
	return true;
}

static int sparc64_verify_line_number(unsigned long pc, unsigned long low,
		unsigned long high)
{
	return true;
}

static int sparc64_dis_filter(ulong vaddr, char *inbuf, unsigned int radix)
{
		return false;
}

static int sparc64_eframe_search(struct bt_info *bt)
{
	struct {struct sparc_stackf sf; struct pt_regs pr;}
		exception_frame_data;
	unsigned long exception_frame = bt->stacktop;
	unsigned long first_frame;
	struct reg_window one_down;
	int rc;

	exception_frame = exception_frame - TRACEREG_SZ - STACKFRAME_SZ;
	rc = readmem(exception_frame, KVADDR, &exception_frame_data,
		sizeof (exception_frame_data), "EF fetch.", RETURN_ON_ERROR);
	if (!rc)
		goto out;
	if (exception_frame_data.pr.magic != 0x57ac6c00)
		goto out;
	first_frame = exception_frame - sizeof (struct reg_window);

	rc = readmem(first_frame, KVADDR, &one_down, sizeof (struct reg_window),
		"Stack fetch.", RETURN_ON_ERROR);
	if (!rc)
		goto out;
	/* Extra arguments. */
	first_frame = first_frame - (6 * 8);

	rc = readmem(first_frame, KVADDR, &one_down, sizeof (struct reg_window),
		"Stack fetch.", RETURN_ON_ERROR);
	if (!rc)
		goto out;
out:
	return rc;
}

/* Need to handle hardirq and softirq stacks. */
static int kstack_valid(struct bt_info *bt, unsigned long sp)
{
	unsigned long thread_info = SIZE(thread_info);
	unsigned long base = bt->stackbase + thread_info;
	unsigned long top = bt->stacktop - sizeof (struct sparc_stackf) -
		sizeof (struct pt_regs);
	int rc = false;

	if (sp & (16U - 1))
		goto out;

	if ((sp >= base) && (sp <= top))
		rc = true;
out:
	return rc;
}

static void sparc64_print_eframe(struct bt_info *bt, unsigned long stack_top)
{
	struct {struct sparc_stackf sf; struct pt_regs pr;} k_entry;
	struct pt_regs *regs = &k_entry.pr;
	unsigned long efp;
	unsigned int tt;
	int rc;

	efp = stack_top - (sizeof (struct pt_regs) +
		sizeof (struct sparc_stackf));
	rc = readmem(efp, KVADDR, &k_entry, sizeof(k_entry),
		"Stack frame and pt_regs.", RETURN_ON_ERROR);
	if (!rc)
		goto out;
	/* Kernel thread or not in kernel any longer? */
	if ((regs->magic & ~MAGIC_TT) != PT_REGS_MAGIC)
		goto out;
	tt = regs->magic & MAGIC_TT;
	fprintf(fp, "TSTATE=0x%lx TT=0x%x TPC=0x%lx TNPC=0x%lx\n",
		regs->tstate, tt, regs->tpc, regs->tnpc);
	fprintf(fp, " r0=0x%.16lx  r1=0x%.16lx  r2=0x%.16lx\n",
		regs->u_regs[0],
		regs->u_regs[1],
		regs->u_regs[2]);
	fprintf(fp, " r3=0x%.16lx  r4=0x%.16lx  r5=0x%.16lx\n",
		regs->u_regs[3],
		regs->u_regs[4],
		regs->u_regs[5]);
#define	___INS	(8)
	fprintf(fp, " r6=0x%.16lx  r7=0x%.16lx r24=0x%.16lx\n",
		regs->u_regs[6],
		regs->u_regs[7],
		regs->u_regs[___INS+0]);
	fprintf(fp, "r25=0x%.16lx r26=0x%.16lx r27=0x%.16lx\n",
		regs->u_regs[___INS+1],
		regs->u_regs[___INS+2],
		regs->u_regs[___INS+3]);
	fprintf(fp, "r28=0x%.16lx r29=0x%.16lx r30=0x%.16lx\n",
		regs->u_regs[___INS+4],
		regs->u_regs[___INS+5],
		regs->u_regs[___INS+6]);
	fprintf(fp, "r31=0x%.16lx\n",
		regs->u_regs[___INS+7]);
#undef	___INS
out:
	return;
}

static void sparc64_print_frame(struct bt_info *bt, int cnt, unsigned long ip,
						unsigned long ksp)
{
	char *symbol = closest_symbol(ip);

	fprintf(fp, "#%d [%lx] %s at %lx\n", cnt, ksp, symbol, ip);

	if (bt->flags & BT_LINE_NUMBERS) {
		char buf[BUFSIZE];

		get_line_number(ip, buf, false);
		if (strlen(buf))
			fprintf(fp, "\t%s\n", buf);
	}
}

static void sparc64_back_trace(struct bt_info *bt)
{
	unsigned long stack_top = bt->stacktop;
	unsigned long ip = bt->instptr;
	unsigned long ksp = bt->stkptr;
	struct reg_window window;
	int cnt = 0;
	int rc;

	do {
		if (!kstack_valid(bt, ksp + STACK_BIAS))
			break;
		rc = readmem(ksp + STACK_BIAS, KVADDR, &window, sizeof(window),
			"KSP window fetch.", RETURN_ON_ERROR);
		if (!rc)
			goto out;
		sparc64_print_frame(bt, cnt, ip, ksp);
		ksp = window.ins[6];
		ip = window.ins[7];
		cnt++;
	} while(cnt != 50);
	sparc64_print_eframe(bt, stack_top);
out:
	return;
}

static ulong sparc64_processor_speed(void)
{
	int cpu;
	unsigned long clock_tick;
	struct syment *sp;

	if (!MEMBER_EXISTS("cpuinfo_sparc", "clock_tick")) {
		error(WARNING, "sparc64 expects clock_tick\n");
		return 0UL;
	}

	sp = per_cpu_symbol_search("__cpu_data");
	if (!sp)
		return 0UL;
	for (cpu = 0; cpu < kt->cpus; cpu++) {
		if (!in_cpu_map(ONLINE, cpu))
			continue;
		if (!readmem(sp->value + kt->__per_cpu_offset[cpu] +
			     MEMBER_OFFSET("cpuinfo_sparc", "clock_tick"),
			     KVADDR, &clock_tick, sizeof(clock_tick),
			     "clock_tick", QUIET|RETURN_ON_ERROR))
			continue;
		return clock_tick/1000000;
	}
	return 0UL;
}

static ulong sparc64_get_task_pgd(ulong task)
{
	struct task_context *tc = task_to_context(task);
	ulong pgd = NO_TASK;

	if (!tc)
		goto out;
	readmem(tc->mm_struct + OFFSET(mm_struct_pgd), KVADDR,
		&pgd, sizeof (unsigned long), "User pgd.", RETURN_ON_ERROR);
out:
	return pgd;
}

static int sparc64_uvtop(struct task_context *tc, ulong va, physaddr_t *ppaddr,
						int verbose)
{
	unsigned long pgd = sparc64_get_task_pgd(tc->task);
	unsigned long paddr;
	int rc = false;

	if (pgd == NO_TASK)
		goto out;
	paddr = sparc64_page_table_walk(pgd, va, verbose);
	/* For now not_valid_pte skips checking for swap pte. */
	if (paddr == not_valid_pte) {
		*ppaddr = 0UL;
		goto out;
	}
	*ppaddr = paddr;
	rc = true;
out:
	return rc;
}

static unsigned long sparc64_vmalloc_translate(unsigned long vaddr, int verbose)
{
	unsigned long paddr = sparc64_page_table_walk(vt->kernel_pgd[0],
							vaddr, verbose);

	return paddr;
}

static unsigned long sparc64_linear_translate(unsigned long vaddr)
{
	unsigned long paddr = __pa(vaddr);

	if (sparc64_verify_paddr(paddr) == false)
		error(FATAL,
			"sparc64_linear_translate: This physical address"
			" (0x%lx) is invalid.", paddr);

	return paddr;
}

static int sparc64_is_vmalloc_mapped(unsigned long vaddr)
{
	struct machine_specific *ms = &sparc64_machine_specific;
	int rc = 0;

	/* We exclude OBP range whose TTEs were captured by kernel in
	 * early boot. It is possible to fetch these but for what purpose?
	 */
	if ((vaddr >= ms->modules_vaddr && vaddr < ms->modules_end) ||
		(vaddr >= ms->vmalloc_start && vaddr < ms->vmalloc_end))
		rc = 1;
	return rc;
}

static int sparc64_is_kvaddr(ulong vaddr)
{
	int rc = kimage_va_range(vaddr);

	if (rc)
		goto out;
	rc = sparc64_is_linear_mapped(vaddr);
	if (rc)
		goto out;
	rc = sparc64_is_vmalloc_mapped(vaddr);
	if (rc)
		goto out;
	rc = sparc64_is_vmemmap_mapped(vaddr);
out:
	return rc;
}

static int sparc64_kvtop(struct task_context *tc, ulong vaddr,
					physaddr_t *paddr, int verbose)
{
	unsigned long phys_addr;
	int rc = false;

	if (kimage_va_range(vaddr))
		phys_addr = kimage_va_translate(vaddr);
	else if (sparc64_is_vmalloc_mapped(vaddr)) {
		phys_addr = sparc64_vmalloc_translate(vaddr, verbose);
		if (phys_addr == not_valid_pte)
			goto out;
	} else if (sparc64_is_linear_mapped(vaddr))
		phys_addr = sparc64_linear_translate(vaddr);
	else if (sparc64_is_vmemmap_mapped(vaddr))
		phys_addr = sparc64_vmemmap_translate(vaddr);
	else {
		error(WARNING,
		"This is an invalid kernel virtual address=0x%lx.",
			vaddr);
		goto out;
	}

	*paddr = phys_addr;
	rc = true;
out:
	return rc;
}

static int sparc64_is_task_addr(ulong task)
{
	int rc = false;
	int cpu;

	if (sparc64_is_linear_mapped(task) || kimage_va_range(task))
			rc = true;
	else {
		for (cpu = 0; cpu < kt->cpus; cpu++)
			if (task == tt->idle_threads[cpu]) {
				rc = true;
				break;
			}
	}
	return rc;
}

static int sparc64_is_uvaddr(ulong vaddr, struct task_context *tc)
{
	struct machine_specific *ms = &sparc64_machine_specific;
	int rc = false;

	if (vaddr < ms->userspace_top)
		rc = true;
	return rc;
}

static const char *pte_page_size(unsigned long pte)
{
	static const char *_4Mb = "4Mb";
	static const char *_64Kb = "64Kb";
	static const char *_8Kb = "8Kb";
	static const char *_ns = "Not Supported";
	const char *result;

	switch (pte & _PAGE_SZALL_4V) {
	case _PAGE_SZ8K_4V:
		result = _8Kb;
		break;
	case _PAGE_SZ64K_4V:
		result = _64Kb;
		break;
	case _PAGE_SZ4MB_4V:
		result = _4Mb;
		break;
	default:
		result = _ns;
	}
	return result;
}

static int sparc64_translate_pte(unsigned long pte, void *physaddr,
							ulonglong unused)
{
	unsigned long paddr = pte_to_pa(pte);
	int rc = false;
	int cnt = 0;

	/* Once again not handling swap pte.*/
	if ((pte & _PAGE_VALID) == 0UL)
		goto out;
	if (pte & _PAGE_NFO_4V)
		fprintf(fp, "%sNoFaultOn", cnt++ ? "|" : "");
	if (pte & _PAGE_MODIFIED_4V)
		fprintf(fp, "%sModified", cnt++ ? "|" : "");
	if (pte & _PAGE_ACCESSED_4V)
		fprintf(fp, "%sAccessed", cnt++ ? "|" : "");
	if (pte & _PAGE_READ_4V)
		fprintf(fp, "%sReadSoftware", cnt++ ? "|" : "");
	if (pte & _PAGE_WRITE_4V)
		fprintf(fp, "%sWriteSoftware", cnt++ ? "|" : "");
	if (pte & _PAGE_P_4V)
		fprintf(fp, "%sPriv", cnt++ ? "|" : "");
	if (pte & _PAGE_EXEC_4V)
		fprintf(fp, "%sExecute", cnt++ ? "|" : "");
	if (pte & _PAGE_W_4V)
		fprintf(fp, "%sWritable", cnt++ ? "|" : "");
	if (pte & _PAGE_PRESENT_4V)
		fprintf(fp, "%sPresent", cnt++ ? "|" : "");
	fprintf(fp, "|PageSize(%s)\n", pte_page_size(pte));
	if (physaddr)
		* (unsigned long *) physaddr = paddr;
	rc = true;
out:
	return rc;
}

static void sparc64_get_frame(struct bt_info *bt, unsigned long *r14,
					unsigned long *r15)
{
	unsigned long ksp_offset = sparc64_ksp_offset + bt->tc->thread_info;
	unsigned long ksp;
	int rc;

	/* We need thread_info's ksp. This is the stack for sleeping threads
	 * and captured during switch_to. The rest is fetchable from there.
	 */
	rc = readmem(ksp_offset, KVADDR, &ksp, sizeof (ksp), "KSP Fetch.",
		RETURN_ON_ERROR);
	if (!rc)
		goto out;
	*r14 = ksp;
	*r15 = symbol_value("switch_to_pc");
out:
	return;
}

static void sparc64_get_dumpfile_stack_frame(struct bt_info *bt,
					unsigned long *psp, unsigned long *ppc)
{
	struct pt_regs *pt_regs;

	/* Is there a better way to get the stack pointer? */
	sparc64_get_frame(bt, psp, ppc);

	pt_regs = (struct pt_regs *)bt->machdep;

	if (!pt_regs)
		fprintf(fp, "0%lx: registers not saved\n", bt->task);

	/* pt_regs can be unaligned */
	BCOPY(ppc, &pt_regs->tpc, sizeof(ulong));
}

static void sparc64_get_stack_frame(struct bt_info *bt, unsigned long *pcp,
						unsigned long *psp)
{
	unsigned long r14, r15;

	if (DUMPFILE() && is_task_active(bt->task))
		sparc64_get_dumpfile_stack_frame(bt, &r14, &r15);
	else
		sparc64_get_frame(bt, &r14, &r15);
	if (pcp)
		*pcp = r15;
	if (psp)
		*psp = r14;
}

static int sparc64_get_kvaddr_ranges(struct vaddr_range *vrp)
{
	struct machine_specific *ms = machdep->machspec;

	vrp[0].type = KVADDR_UNITY_MAP;
	vrp[0].start = ms->page_offset;
	vrp[0].end = ~0ULL;
	vrp[1].type = KVADDR_VMALLOC;
	vrp[1].start = ms->vmalloc_start;
	vrp[1].end = ms->vmalloc_end;
	vrp[2].type = KVADDR_VMEMMAP;
	vrp[2].start = ms->vmemmap_vaddr;
	vrp[2].end = ms->vmemmap_end;
	/* WTF is this? */
	vrp[3].type = KVADDR_START_MAP;
	vrp[3].start = symbol_value("_start");
	vrp[3].end = symbol_value("_end");
	vrp[4].type = KVADDR_MODULES;
	vrp[4].start = ms->modules_vaddr;
	vrp[4].end = ms->modules_end;
	return MAX_KVADDR_RANGES;
}

static void sparc64_get_crash_notes(void)
{
	unsigned long *notes_ptrs, size, crash_notes_address;
	int ret;

	if (!symbol_exists("crash_notes")) {
		error(WARNING, "Could not retrieve crash_notes.");
		goto out;
	}

	crash_notes_address = symbol_value("crash_notes");
	size = kt->cpus * sizeof(notes_ptrs[0]);
	notes_ptrs = (unsigned long *) GETBUF(size);
	ret = readmem(crash_notes_address, KVADDR, notes_ptrs, size,
		"crash_notes", RETURN_ON_ERROR);
	if (!ret)
		goto out2;
out2:
	FREEBUF(notes_ptrs);
out:
	return;
}

static void sparc64_init_kstack_info(void)
{
	sparc64_ksp_offset = MEMBER_OFFSET("thread_info", "ksp");
}

static void sparc64_init_irq_stacks(void)
{
	void *irq_stack;
	unsigned long stack_size;

	stack_size = get_array_length("hardirq_stack", NULL, 0) *
		     sizeof(unsigned long);
	irq_stack = malloc(stack_size);
	if (!irq_stack)
		error(FATAL, "malloc failure in sparc64_init_irq_stacks");

	get_symbol_data("hardirq_stack", stack_size, irq_stack);
	tt->hardirq_ctx = irq_stack;

	stack_size = get_array_length("softirq_stack", NULL, 0) *
		     sizeof(unsigned long);
	irq_stack = malloc(stack_size);
	if (!irq_stack)
		error(FATAL, "malloc failure in sparc64_init_irq_stacks");

	get_symbol_data("softirq_stack", stack_size, irq_stack);
	tt->softirq_ctx = irq_stack;
}

static void sparc64_init_vmemmap_info(void)
{
	struct machine_specific *ms = &sparc64_machine_specific;
	unsigned long page_struct_size;

	if (!symbol_exists("vmemmap_table"))
		return;
	vmemmap_table = symbol_value("vmemmap_table");
	if (get_array_length("vmemmap_table", NULL, 0) == 256)
		two_level_vmemmap_table = true;
	page_struct_size = STRUCT_SIZE("page");
	ms->vmemmap_vaddr = SPARC64_VMEMMAP_BASE;
	ms->vmemmap_end = SPARC64_VMEMMAP_BASE +
		((1UL << (machdep->max_physmem_bits - PAGE_SHIFT)) *
		 page_struct_size);
}

static void sparc64_init_cpu_info(void)
{
	unsigned long trap_block, per_cpu_base_offset, per_cpu_base;
	unsigned long trap_per_cpu;
	int cpu;

	if (!symbol_exists("trap_block"))
		error(FATAL, "sparc64 requires trap_block symbol.\n");

	trap_block = symbol_value("trap_block");
	if (!MEMBER_EXISTS("trap_per_cpu", "__per_cpu_base"))
		error(FATAL, "sparc64 requires __per_cpu_base.\n");
	trap_per_cpu = STRUCT_SIZE("trap_per_cpu");
	per_cpu_base_offset = MEMBER_OFFSET("trap_per_cpu", "__per_cpu_base");
	for (cpu = 0; cpu < NR_CPUS; cpu++,
		trap_block = trap_block + trap_per_cpu) {

		if (!in_cpu_map(POSSIBLE, cpu))
			continue;
		readmem(trap_block + per_cpu_base_offset, KVADDR,
			&per_cpu_base, sizeof(per_cpu_base),
			"sparc64: per_cpu_base", FAULT_ON_ERROR);
		kt->__per_cpu_offset[cpu] = per_cpu_base;
	}
}

void sparc64_init(int when)
{
	struct machine_specific *ms = &sparc64_machine_specific;

	switch (when)
	{
	case SETUP_ENV:
		machdep->process_elf_notes = process_elf64_notes;
		break;
	case PRE_SYMTAB:
                machdep->machspec = ms;
		machdep->verify_paddr = sparc64_verify_paddr;
		machdep->verify_symbol = sparc64_verify_symbol;
		machdep->verify_line_number = sparc64_verify_line_number;

                if (pc->flags & KERNEL_DEBUG_QUERY)
                        return;
		machdep->flags |= MACHDEP_BT_TEXT;
                if (machdep->cmdline_args[0])
                        sparc64_parse_cmdline_args();
		break;

	case PRE_GDB:
		if (kernel_symbol_exists("swapper_pud_dir"))
			sparc64_init_uek3();
		else if (kernel_symbol_exists("vmemmap_table"))
			sparc64_init_level3();
		else
			sparc64_init_level4();

		machdep->pagesize = memory_page_size();
		machdep->pageshift = ffs(machdep->pagesize) - 1;
		machdep->pageoffset = machdep->pagesize - 1;
		machdep->pagemask = ~((ulonglong) machdep->pageoffset);
		machdep->stacksize = machdep->pagesize * 2;

	        machdep->eframe_search = sparc64_eframe_search;
	        machdep->back_trace = sparc64_back_trace;
	        machdep->processor_speed = sparc64_processor_speed;

	        machdep->uvtop = sparc64_uvtop;
	        machdep->kvtop = sparc64_kvtop;
	        machdep->get_task_pgd = sparc64_get_task_pgd;

		machdep->dump_irq = generic_dump_irq;

		machdep->get_stack_frame = sparc64_get_stack_frame;
		machdep->get_stackbase = generic_get_stackbase;
		machdep->get_stacktop = generic_get_stacktop;
		machdep->translate_pte = sparc64_translate_pte;
		machdep->memory_size = generic_memory_size;

		machdep->vmalloc_start = sparc64_vmalloc_start;
		machdep->is_task_addr = sparc64_is_task_addr;
                machdep->is_kvaddr = sparc64_is_kvaddr;
                machdep->is_uvaddr = sparc64_is_uvaddr;
		machdep->dis_filter = sparc64_dis_filter;
		machdep->get_smp_cpus = sparc64_get_smp_cpus;
		machdep->clear_machdep_cache = sparc64_clear_machdep_cache;
		machdep->get_kvaddr_ranges = sparc64_get_kvaddr_ranges;
		machdep->cmd_mach = sparc64_cmd_mach;
		machdep->init_kernel_pgd = sparc64_init_kernel_pgd;
		machdep->value_to_symbol = generic_machdep_value_to_symbol;
		machdep->show_interrupts = generic_show_interrupts;

		pt_level_alloc(&machdep->pgd, "Can't malloc pgd space.");
		pt_level_alloc(&machdep->pud, "Can't malloc pud space.");
		pt_level_alloc(&machdep->pmd, "Can't malloc pmd space.");
		pt_level_alloc(&machdep->ptbl, "Can't malloc ptbl space.");
		pt_clear_cache();
		if (symbol_exists("vmemmap_populate"))
			machdep->flags |= VMEMMAP;
		sparc64_phys_base();
		sparc64_kimage_limits();
		break;

	case POST_GDB:
		if (kernel_symbol_exists("PAGE_OFFSET"))
			get_symbol_data("PAGE_OFFSET", sizeof(unsigned long),
					&machdep->machspec->page_offset);
		else
			machdep->machspec->page_offset = SPARC64_PAGE_OFFSET;
	        machdep->kvbase = symbol_value("_stext");
		machdep->identity_map_base = (ulong) PAGE_OFFSET;
		machdep->ptrs_per_pgd = PTRS_PER_PGD;
		ms->modules_vaddr = SPARC64_MODULES_VADDR;
		ms->modules_end = SPARC64_MODULES_END;
		ms->vmalloc_start = SPARC64_VMALLOC_START;
		ms->vmalloc_end = SPARC64_VMALLOC_END;
		machdep->section_size_bits = _SECTION_SIZE_BITS;
		if (kernel_symbol_exists("nr_irqs"))
			get_symbol_data("nr_irqs", sizeof(unsigned int),
			&machdep->nr_irqs);
		sparc64_init_vmemmap_info();
		sparc64_init_cpu_info();
		sparc64_init_kstack_info();
		sparc64_init_irq_stacks();
		break;
	case POST_VM:
		if (!ACTIVE())
			sparc64_get_crash_notes();
		break;
	case POST_INIT:
		break;

	case LOG_ONLY:
                machdep->machspec = &sparc64_machine_specific;
		machdep->kvbase = kt->vmcoreinfo._stext_SYMBOL;
		break;
	}
}

void sparc64_dump_machdep_table(ulong unused)
{
}

#endif /* SPARC64 */
